#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>

struct node {
    char * in;
    char * out;
};

int execNode (struct node * parm) {
struct hostent * he;
unsigned long ip0;
int retval;
char * v;
int l;

    if (parm == NULL || parm->in == NULL || parm->out != NULL) {
        fprintf (stderr, "Internal error [%d]\n", __LINE__);
        parm->out = strdup ("error");
        return -__LINE__;
    }

    ip0 = inet_addr (parm->in);
    if ((ip0 != 0) && (ip0 != ~0)) {
        /* This operation may take some time */
        he = gethostbyaddr ((char *)&ip0, 4, AF_INET);
    } else {
        he = NULL;
    }

    if ((he != NULL) && (he->h_name != NULL) && (he->h_name[0] != '\0')) {
        retval = 0;
        v = he->h_name;
    } else {
        retval = 1;
        v = parm->in;
    }

    v = strdup (v);
    l = strlen (v);

    /* Remove trailing CR/LFs */
    while (l > 0) {
        l--;
        if ((v[l] == 10) || (v[l] == 13)) {
            v[l] = '\0';
        } else break;
    }

    /* Other non-ASCII characters are substituted with ? */
    while (l >= 0) {
        if (v[l] != '\0' && ((v[l] < 31) || ((unsigned char)v[l] > 127))) {
            v[l] = '?';
        }
        l--;
    }

    if (v[0] == '\0') {
        free (v);
        parm->out = strdup ("error");
        fprintf (stderr, "\07"); fflush (stdout);
    }

    /* This has to be last step */
    parm->out = v;
    return retval;
}

void atomicAdd (int * a, int b) {
    _asm {
        mov eax, a
        mov ebx, b
        add [eax], ebx
    }
}

static int launchedThreadCount = 0;
static int activeThreadCount = 0;

#define MAX_THREADS_LAUCHED (256)
#define MAX_THREADS_ACTIVE (256)

static void asyncExecNode (void * vparm) {

    /* Not a technically correct counting semaphore implementation but will
       work correctly in all practical situations */
    WaitToStartLoop:;
    while (activeThreadCount >= MAX_THREADS_ACTIVE) {
        Sleep (0);
    }
    atomicAdd (&activeThreadCount, 1);
    if (activeThreadCount > MAX_THREADS_ACTIVE) {
        atomicAdd (&activeThreadCount, -1);
        goto WaitToStartLoop;
    }

    execNode ((struct node *)vparm);
    atomicAdd (&activeThreadCount, -1);
    atomicAdd (&launchedThreadCount, -1);
    _endthread ();
}

struct wnode {
    struct node parm;
    struct wnode * next;
};

static void launchThreads (void * top) {
struct wnode * next;

    /* Launch threads to consume the parameters */
    next = (struct wnode *)top;
    while (next != NULL) {
        while (launchedThreadCount >= MAX_THREADS_LAUCHED) {
            Sleep (0);
        }
        if (!_beginthread (asyncExecNode, 4096, (void *)&(next->parm))) {
            execNode (&(next->parm));
        } else {
            atomicAdd (&launchedThreadCount, 1);
        }
        next = next->next;
    }
    _endthread ();
}

int mt (char * filename) {
FILE * fp;
char line[64];
struct wnode * top;
struct wnode ** next;

    fp = fopen (filename, "rt");
    if (fp == NULL) {
        fprintf (stderr, "Unable to open \"%s\"\n", filename);
        return -__LINE__;
    }

    /* Build list of parameter nodes */
    next = &top;
    while (fgets (line, 63, fp)) {
        *next = (struct wnode *)malloc (sizeof (struct wnode));
        (*next)->parm.in  = strdup (line);
        (*next)->parm.out = NULL;
        (*next)->next     = NULL;
        next = &((*next)->next);
    }
    fclose (fp);

    /* Launch threads to consume the parameters */
    _beginthread (launchThreads, 4096, (void *)top);

    /* Dump then remove parameter nodes */
    while (top != NULL) {
        struct wnode * tmp;

        while (top->parm.out == NULL) {
            Sleep (0);
        }
        printf ("%s\n", top->parm.out);

        if (top->parm.in  != NULL) free (top->parm.in);
        if (top->parm.out != NULL) free (top->parm.out);
        tmp = top;
        top = top->next;
        free (tmp);
    }

    return 0;
}

int main (int argc, char *argv[]) {
struct node parm;
int retval;
WSADATA winsockData;
WORD versionRequired = 0x0101; /* Version 1.1 */

    if (argc < 2) {
        fprintf (stderr, "Syntax: %s [ip-address]\n", argv[0]);
        fprintf (stderr, "     or %s @[ip-address-file]", argv[0]);
        return -1;
    }

    retval = WSAStartup (versionRequired, &winsockData);
    if (retval != 0) {
        fprintf (stderr, "Internal error [%d] Unable to start winsock", __LINE__);
        return -__LINE__;
    }

    if (argv[1][0] == '@') {
        if (argv[1][1] == '\0' && argc > 2) {
           retval = mt (argv[2]);
        } else {
           retval = mt (argv[1] + 1);
        }
    } else {
        parm.in  = argv[1];
        parm.out = NULL;
        retval = execNode (&parm);
        if (parm.out) printf ("%s", parm.out);
    }

    WSACleanup ();
    return retval;
}

